/* ----------------------------------------------------------------------------
 * $Id: niswatchd.c,v 1.1 2004/06/30 13:22:51 stephan Exp $
 * ----------------------------------------------------------------------------
 * niswatchd: watch the nisdomain directory for updates and if one occurs,
 *            have a kicker application perform some actions  (like rebuilding
 *            generated mail alias files or checking home directories)
 * ----------------------------------------------------------------------------
 * $Log: niswatchd.c,v $
 * Revision 1.1  2004/06/30 13:22:51  stephan
 * Initial revision
 *
 * ----------------------------------------------------------------------------
 * Copyright (C) 2004 Stephan Leemburg <stephan@jvc.nl> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */
#include <sys/types.h>

#ifdef __APPLE__ /* compiles, not tested however */
#define KQUEUE
#endif
#ifdef __FreeBSD__
#define KQUEUE
#endif
#ifdef __linux__
#define DNOTIFY
#endif

#ifdef KQUEUE	/* *BSD */
#include <sys/event.h>
#endif

#ifdef DNOTIFY /* Linux */
#define __USE_GNU
#endif

#include <fcntl.h>

#ifdef DNOTIFY /* Linux */
#undef __USE_GNU
#endif

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <dirent.h>
#include <syslog.h>

#define MYNAME		"niswatchd"
#define MYVERSION	"0.1"
#define PIDFMT		"/var/run/" MYNAME "-%s.pid"
#define NISBASE		"/var/yp"
#define NISKICKER	"/usr/local/bin/niskicker"
#define PIDLEN		128 /* string size of pid */

#define STARTING	0
#define RUNNING		1
#define STOPPING	2

#define TIMER		1800 /* logmark every 15 minutes */
#define MAX_PENDING	5
#define SECSTOLINGER	0
#define NSECSTOLINGER	500000000 /* nano -> 10^-9, micro 10^-6 */

static int mylock = -1;
static char pidfile[FILENAME_MAX+1] = { 0 };
static volatile int mystate = STARTING;

static int monitor(char*, char*, char*);
static int kick(char*, char*, char*);
static int watch(int);

#ifdef KQUEUE
static int kqueuewatch(int);
#endif
#ifdef DNOTIFY
static int dnotifywatch(int);
#endif

static void sighandler(int);
#ifdef DNOTIFY
static void sigiohandler(int);
#endif
static void setuphandler(int, void (*handler)(int));
static int startservice(char*);
static int stopservice(int);
static void usage(char*, char *, char*, char*, char*);

int main(int argc, char **argv)
{
	char *base = NISBASE;
	char *kicker = NISKICKER;
	char domain[MAXHOSTNAMELEN+1] = { 0 };
	struct stat sb;
	int c;

	openlog(MYNAME, LOG_CONS, LOG_DAEMON);

	if(geteuid())
	{
		syslog(LOG_ERR, "unfortunatly you have to be root.");
		return EXIT_FAILURE;
	}

	if(!domain[0] && getdomainname(domain, sizeof(domain)-1))
	{
		syslog(LOG_ERR, "getdomainname: %m");
		return EXIT_FAILURE;
	}

	while((c=getopt(argc, argv, "b:d:k:")) >= 0)
		switch(c)
		{
		case 'b':
			base = optarg;
			break;
		case 'd':
			snprintf(domain, sizeof(domain), "%s", optarg);
			break;
		case 'k':
			kicker = optarg;
			break;
		default:
			usage(MYNAME, MYVERSION, NISBASE, domain, NISKICKER);
			return EXIT_FAILURE;
		}

	argc -= optind;

	if(argc)
		syslog(LOG_ERR, "stray arguments on commandline (ignoring)");

	if(!domain[0])
	{
		syslog(LOG_ERR, "domainname empty");
		return EXIT_FAILURE;
	}

	if(stat(kicker, &sb))
	{
		syslog(LOG_ERR, "cannot stat(%s) (%m)", kicker);
		return EXIT_FAILURE;
	}
	
	if(sb.st_uid != geteuid())
	{
		syslog(LOG_ERR, "kicker(%s) is not owned by %d", 
			kicker, geteuid());
		return EXIT_FAILURE;
	}

	if(!(sb.st_mode & S_IFREG) || !(sb.st_mode & 0100))
	{
		syslog(LOG_ERR, "kicker(%s) is not executable", kicker);
		return EXIT_FAILURE;
	}

	if(sb.st_mode & 0022)
	{
		syslog(LOG_ERR, "kicker(%s) is writeable by group/others", 
			kicker);
		return EXIT_FAILURE;
	}

	if(startservice(&domain[0]) != EXIT_SUCCESS)
		return EXIT_FAILURE;

	return stopservice(monitor(base, domain, kicker));
}


static int monitor(char *base, char *domain, char *kicker)
{
	int fd = -1;
	char path[FILENAME_MAX+1];

	syslog(LOG_INFO, "monitoring domain %s within %s with kicker %s", 
		domain, base, kicker);

	snprintf(path, sizeof(path), "%s/%s", base, domain);

	fd = open(path, O_RDONLY);
	if(fd < 0)
	{
		syslog(LOG_ERR, "cannot open '%s' (%m)", path);
		return EXIT_FAILURE;
	}
	
	while(mystate != STOPPING)
	{
		if(watch(fd) == EXIT_FAILURE)
			continue;
		
		/* start kicker on change */
	
		if(kick(base, domain, kicker) != EXIT_SUCCESS)
			break;
	}
	if(fd >= 0)
		close(fd);

	return mystate == STOPPING ? EXIT_SUCCESS : EXIT_FAILURE;
}

static int kick(char *base, char *domain, char *kicker)
{
	pid_t pid;
	int status;
	char *service = strrchr(base, '/');
	if(!*service)
		service = base;

	syslog(LOG_INFO, "kicking '%s' on domain '%s'", kicker, domain);

	if((pid = fork()) < 0)
	{
		syslog(LOG_ERR, "cannot fork (%m)");
		return EXIT_FAILURE;
	}
	
	if(!pid) /* child */
	{
		execl(kicker, service, base, domain, NULL);
		syslog(LOG_ERR, "exec of kicker '%s' failed (%m)", kicker);
		_exit(EXIT_FAILURE);
	} 
	(void)waitpid(pid, &status, 0);

	if(WIFEXITED(status))
		syslog(LOG_INFO, "kicker exited with status (%d)", 
			WEXITSTATUS(status));
	else if(WIFSIGNALED(status))
		syslog(LOG_ERR, "kicker terminated with signal (%d)", 
			WTERMSIG(status));
	else
		syslog(LOG_ERR, "kicker seems to have stopped");

	return EXIT_SUCCESS; /* continue to monitor */
}

static int watch(int fd)
{
#ifdef KQUEUE
	return kqueuewatch(fd);
#endif
#ifdef DNOTIFY
	return dnotifywatch(fd);
#endif
}

#ifdef KQUEUE /* *BSD */
static int kqueuewatch(int fd)
{
	int fired;
	static int kq = -1;
	struct kevent watch;
	struct kevent event;
	struct timespec timer;

	bzero(&watch, sizeof(struct kevent));
	bzero(&event, sizeof(struct kevent));

	watch.ident = fd;
	watch.filter = EVFILT_VNODE;
	watch.flags = EV_ADD | EV_CLEAR;
	watch.fflags = NOTE_DELETE|NOTE_WRITE|NOTE_EXTEND|NOTE_ATTRIB|NOTE_LINK;
	watch.data = 0;
	watch.udata = 0;

	if(kq < 0)
		kq = kqueue();

	if(kq < 0) 
	{
		syslog(LOG_ERR, "cannot create kqueue: (%m)");
		return EXIT_FAILURE;
	}
	kevent(kq, &watch, 1, NULL, 0, NULL);

	timer.tv_sec = TIMER;
	timer.tv_nsec = 0;

	for(fired=0;;) 
	{
		int rc = kevent(kq, NULL, 0, &event, 1, &timer);
		if(rc < 0)
		{
			syslog(LOG_ERR, "kevent error: (%m)");
			return EXIT_FAILURE;
		}

		if(rc == 0) 
		{
			/* and kick in if event(s) occured */
			if(fired)
				break;

			timer.tv_sec = TIMER;
			timer.tv_nsec = 0;
			syslog(LOG_NOTICE, "Still watching.");
			continue;
		}
		if(++fired > MAX_PENDING)
			break;

		/* wait a second for another event */

		timer.tv_sec = SECSTOLINGER;
		timer.tv_nsec = NSECSTOLINGER;
		fired++;
	}
	syslog(LOG_INFO, "kicking in for pending %d events", fired);

	return EXIT_SUCCESS;
}
#endif

#ifdef DNOTIFY /* Linux */
static int dnotifywatch(int fd)
{
	int fired;
	struct timeval timer;

	timer.tv_sec = TIMER;
	timer.tv_usec = 0;

	if(fcntl(fd, F_NOTIFY, DN_MODIFY|DN_CREATE)) /* ONESHOT */
	{
		syslog(LOG_ERR, "cannot set DNOTIFY (%m)");
		return EXIT_FAILURE;
	}

	for(fired=0;;) 
	{
		int rc = select(0, 0, 0, 0, &timer);
		if(!rc) /* timer expired */
		{
			/* and kick in if event(s) occured */
			if(fired)
				break;

			timer.tv_sec = TIMER;
			timer.tv_usec = 0;
			syslog(LOG_NOTICE, "Still watching.");
			continue;
		}
		if(++fired > MAX_PENDING)
			break;

		/* wait a second for another event */

		timer.tv_sec = SECSTOLINGER;
		timer.tv_usec = (NSECSTOLINGER / 1000); /* 10^-9 to 10^-6 */
		fired++;
	}
	syslog(LOG_INFO, "kicking in for pending %d events", fired);

	return EXIT_SUCCESS;
}
#endif

#ifdef DNOTIFY
static void sigiohandler(int sig)
{
	;
}
#endif

static void sighandler(int sig)
{
	switch(sig)
	{
	default:
		mystate = STOPPING;
		break;
	}
}

static void setuphandler(int sig, void (*handler)(int))
{
	struct sigaction sa;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	sigaction(sig, &sa, 0);
}

static int startservice(char *domain)
{
	char pid[PIDLEN];
	struct flock sf;

	snprintf(pidfile, FILENAME_MAX, PIDFMT, domain);

	mylock = open(pidfile, O_RDWR|O_CREAT, 0644);
	if(mylock < 0)
	{
		syslog(LOG_ERR, "cannot open(%s) (%m)", pidfile);
		return EXIT_FAILURE;
	}

	if(daemon(0, 0))
	{
		syslog(LOG_ERR, "cannot daemonize (%m)");
		return EXIT_FAILURE;
	}

	sf.l_type = F_WRLCK;
	sf.l_whence = sf.l_start = sf.l_len = 0;
	sf.l_pid = 0;

	if(fcntl(mylock, F_SETLK, &sf) < 0)
	{
		syslog(LOG_ERR, "cannot lock(" MYNAME ") (%m)");
		return EXIT_FAILURE;
	}

	snprintf(pid, sizeof(pid), "%d\n", getpid());

	if(ftruncate(mylock, 0) 
	|| write(mylock, pid, strlen(pid)) != strlen(pid))
	{
		syslog(LOG_ERR, "cannot write(" MYNAME ") (%m)");
		return EXIT_FAILURE;
	}

	mystate = RUNNING;

	setuphandler(SIGINT, &sighandler);
	setuphandler(SIGQUIT, &sighandler);
	setuphandler(SIGTERM, &sighandler);
#ifdef DNOTIFY
	setuphandler(SIGIO, &sigiohandler);
#endif

	return EXIT_SUCCESS;
}

static int stopservice(int status)
{
	syslog(LOG_INFO, "stopping service " MYNAME " %s" ,
		status == EXIT_FAILURE ? "with an error" : "normally");

	if(mylock >= 0)
	{
		unlink(pidfile);
		close(mylock); /* and release fcntl lock */
	}
	return status;
}

static void usage(char *name, char *vs, char *base, char *domain, char *kicker)
{
	fprintf(stderr, "usage: %s [-b <base>] [-d <domain>] [-k <kicker>]\n"
		"where:\n"
		" -b <base>   : use <base> as the nis repository basename\n"
		"             : default (%s)\n"
		" -d <domain> : use <domain> as the nis domainname\n"
		"             : default (%s)\n"
		" -k <kicker> : use <kicker> as the change kicker\n"
		"             : default (%s)\n"
		"Version: %s\n",
		name,
		base,
		domain,
		kicker,
		vs
	);
}
